#!/usr/bin/env python3

from copy import copy
from collections.abc import Mapping, Iterable
from functools import partial
from unittest.mock import Mock

SENTINEL = object()

class Namespace(Mapping):
    """\
    Evaluate and/or execute snippets of python code, with powerful control over the 
    names available to those snippets.

    .. note::
    
        It is conventional to:

        - Name your namespaces: ``with_{short description}``

        - Create any namespaces you will need in a single helper file (e.g.  
          ``param_helpers.py``) to be imported in test scripts as necessary.  
          Note that namespaces are immutable, so it's safe for them to be 
          global variables.

    Examples:

        The first step when using a namespace is to specify which names it 
        should include.   This can be done using...

        ...strings (which will be `python:exec`'d):
            
            >>> with_math = Namespace('import math')
            >>> with_math.eval('math.sqrt(4)')
            2.0

        ...dictionaries:

            >>> import math
            >>> with_math = Namespace(globals())
            >>> with_math.eval('math.sqrt(4)')
            2.0

        ...modules:
            
            >>> with_math = Namespace(math)
            >>> with_math.eval('math.sqrt(4)')
            2.0

        ...other namespaces (via the constructor):

            >>> with_np = Namespace(with_math, 'import numpy as np')
            >>> with_np.eval('np.arange(3) * math.sqrt(4)')
            array([0., 2., 4.])

        ...other namespaces (via the `fork` method):

            >>> with_np = with_math.fork('import numpy as np')
            >>> with_np.eval('np.arange(3) * math.sqrt(4)')
            array([0., 2., 4.])

        ...keyword arguments:

            >>> with_math = Namespace(math=math)
            >>> with_math.eval('math.sqrt(4)')
            2.0

        ...the `star` function:

            >>> with_math = Namespace(star(math))
            >>> with_math.eval('sqrt(4)')
            2.0

        Once you have an initialized namespace, you can use it to...

        ...evaluate expressions:

            >>> with_math.eval('sqrt(4)')
            2.0
            
        ...execute blocks of code:

            >>> ns = with_math.exec('''
            ... a = sqrt(4)
            ... b = sqrt(9)
            ... ''')
            >>> ns['a']
            2.0
            >>> ns['b']
            3.0
            >>> ns.eval('a + b')
            5.0

        ...make error-detecting context managers:

            >>> class MyError(Exception):
            ...     pass
            ...
            >>> with_err = Namespace(MyError=MyError)
            >>> err_cm = with_err.error({'type': 'MyError', 'pattern': r'\\d+'})
            >>> with err_cm:
            ...    raise MyError('404')
    """

    def __init__(self, *args, **kwargs):
        """
        Construct a namespace containing the given names.

        Arguments:
            args (str,dict,types.ModuleType):
                If string: The string will be executed as python code and any 
                names that are defined by that code will be added to the 
                namespace.  Any names added by previous arguments will be 
                available to the code.

                If dict: The items in the dictionary will the directly added to 
                the namespace.  All of the dictionary keys must be strings.

                If module: The module will be added to the namespace, with its 
                name as the key.

                The *args* are processed in order, so later *args* may 
                overwrite earlier *args*.

            kwargs:
                Each key/value pair will be added to the namespace.  The 
                *kwargs* are processed after the *args*, so if the same key is 
                defined by both, the *kwargs* definition will be used.
        """
        self._dict = {}
        _update_namespace(self._dict, *args, **kwargs)

    def __repr__(self):
        return f'{self.__class__.__name__}({self._dict.__repr__()})'

    def __getitem__(self, key):
        return self._dict.__getitem__(key)

    def __iter__(self):
        return self._dict.__iter__()

    def __len__(self):
        return self._dict.__len__()

    def copy(self):
        """
        Create a shallow copy of this namespace.

        This is an alias for `fork()` with no arguments, for compatibility with 
        the `dict` API.
        """
        return self.fork()

    def fork(self, *args, **kwargs):
        """
        Create a shallow copy of this namespace.

        The arguments allow new names to be added to the new namespace.  All 
        arguments have the same meaning as in the constructor.
        """
        return self.__class__(self, *args, **kwargs)

    def eval(self, *src, keys=False, defer=False):
        """
        Evaluate the given expression within this namespace.

        Arguments:
            src (str,list,dict):
                The expression (or expressions) to evaluate.  Strings are 
                directly evaluated.  List items and dict values are recursively 
                evaluated.  Dict keys are not evaluated unless *keys* is 
                true.  This allows you to switch freely between structured text 
                and python syntax, depending on which makes most sense for each 
                particular input.

            keys (bool):
                If true, evaluate dictionary keys.  Disabled by default because 
                most dictionary keys are strings, and it's annoying to have to 
                quote them.

            defer (bool):
                If true, instead of actually evaluating the given expressions, 
                return a no-argument callable that will evaluate them when 
                called.  The purpose of this is to make it possible to specify 
                that a test parameter should be evaluated in a schema, but to 
                defer actually evaluating it until inside the test function.

        Returns:
            Any: If no *src* expressions were given: A `functools.partial` 
            version of this function that will remember any specified keyword 
            arguments.

            If any *src* expressions were given and the *defer* argument was 
            True: A `functools.partial` version of this function that will 
            remember the expressions and keyword arguments given to it.  The 
            return value will also have an `eval` attribute that aliases the 
            deferred function.

            If any *src* expressions were given and the *defer* argument was 
            False: The result(s) of evaluating the given expression(s).

        Examples:
            Basic usage::

                >>> with_a = Namespace(a=1)
                >>> with_a.eval('a + 1')
                2

            Control whether or not dictionary keys are evaluated::

                >>> with_a.eval({'a': '2'})
                {'a': 2}
                >>> with_a.eval({'a': '2'}, keys=True)
                {1: 2}

            Use a schema to specify that a test parameter should be evaluated, 
            but defer that evaluation until within the test::
                
                >>> @parametrize_from_file(  # doctest: +SKIP
                ...         schema=cast(
                ...             expected=with_a.eval(defer=True),
                ...         ),
                ... )
                ... def test_snippet(expected)
                ...     expected = expected.eval()

        Details:
            `unittest.mock.Mock` instances are handled specially by this 
            method.  Specifically, they are returned unchanged (i.e. without 
            being evaluated).  This special case exists because `error_or` uses 
            `unittest.mock.Mock` instances (by default) as placeholders when an 
            exception is expected.
        """
        if not src:
            return partial(self.eval, keys=keys, defer=defer)
        if defer:
            f = f.eval = partial(self.eval, *src, keys=keys)
            return f

        src = src[0] if len(src) == 1 else list(src)
        recurse = partial(self.eval, keys=keys)

        if type(src) is list:
            return [recurse(x) for x in src]
        elif type(src) is dict:
            f = recurse if keys else lambda x: x
            return {f(k): recurse(v) for k, v in src.items()}
        elif isinstance(src, Mock):
            return src
        else:
            return eval(src, self._dict.copy())

    def exec(self, src=SENTINEL, get=SENTINEL, defer=False):
        """
        Execute the given python snippet within this namespace.

        Arguments:
            src (str): A snippet of python code to execute.

            get (str, collections.abc.Iterable, collections.abc.Callable):
                If string: The name of a variable defined in the given snippet 
                to look up and pass on to the caller.

                If iterable: The names of variables defined in the given 
                snippet to look up and return to the caller.  The return value 
                will be a tuple with the same length as the given iterable.
                
                If callable: The given function will be passed a read-only 
                dictionary containing all the names defined in the given 
                snippet.  Whatever that function returns will be passed on to 
                the caller.

            defer (bool):
                If true, instead of actually evaluating the given snippet, 
                return a no-argument callable that will evaluate it when 
                called.  The purpose of this is to make it possible to specify 
                that a test parameter should be executed in a schema, but to 
                defer actually executing it until inside the test function.

        Returns:
            Any: If *src* was not specified: A `functools.partial` version of 
            this function that will remember any specified keyword arguments.

            If *src* was specified and the *defer* argument was True: A 
            `functools.partial` version of this function that will remember the 
            snippet and keyword arguments given to it.  The return value will 
            also have an `exec` attribute that aliases the deferred function.

            If *src* was specified and *get* was not: A new `Namespace` 
            containing all of the variables defined in the snippet.

            If *src* and *get* were specified: The value(s) of the indicated 
            variable(s) defined the snippet.

        Examples:
            Basic usage::

                >>> with_a = Namespace(a=1)
                >>> with_b = with_a.exec('b = a + 1')
                >>> with_a
                Namespace({'a': 1})
                >>> with_b
                Namespace({'a': 1, 'b': 2})

            Get a specific value from the executed snippet::

                >>> with_a.exec('b = a + 1', get='b')
                2

            Use a schema to specify that a test parameter should be executed, 
            but defer that execution until within the test::
                
                >>> @parametrize_from_file(  # doctest: +SKIP
                ...         schema={
                ...             'snippet': with_a.exec(defer=True),
                ...         },
                ... )
                ... def test_snippet(snippet)
                ...     snippet = snippet.exec()

        Details:
            `unittest.mock.Mock` instances are handled specially by this 
            method.  Specifically, they are returned unchanged (i.e. without 
            being executed).  This special case exists because `error_or` uses
            `unittest.mock.Mock` instances (by default) as placeholders when an 
            exception is expected.
        """
        if src is SENTINEL:
            return partial(self.exec, get=get, defer=defer)
        if defer:
            f = f.exec = partial(self.exec, src, get=get)
            return f
        if isinstance(src, Mock):
            return src

        fork = self.fork()
        exec(src, fork._dict)

        if get is SENTINEL:
            return fork
        elif callable(get):
            return get(fork)
        elif isinstance(get, Iterable) and not isinstance(get, str):
            return tuple(fork[x] for x in get)
        else:
            return fork[get]

    def error(self, exc_spec):
        """\
        Make an exception-checking context manager in the context of this 
        namespace.

        See the `error` documentation for more information.  This method simply 
        calls `error` with the *globals* argument set to this namespace.
        """
        from .schema import error
        return error(exc_spec, globals=self)

    def error_or(self, *expected, **kwargs):
        """
        Make an exception-checking schema function in the context of this 
        namespace.

        See the `error_or` documentation for more information.  This method 
        simply calls `error_or` with the *globals* argument set to this 
        namespace. 
        """
        from .schema import error_or
        return error_or(*expected, **kwargs, globals=self)

def star(module):
    """
    Return a dictionary containing all public attributes exposed by the given 
    module.

    This function follows the same rules as ``from <module> import *``:

    - If the given module defines ``__all__``, only those names will be 
      used.
    - Otherwise, all names without leading underscores will be used.

    The dictionary returned by this function is meant to be used as input to 
    :py:class:`Namespace.__init__()`.
    """
    try:
        keys = module.__all__
    except AttributeError:
        keys = (k for k in module.__dict__ if not k.startswith('_'))

    return {
            k: module.__dict__[k]
            for k in keys
    }

def _update_namespace(ns_dict, *args, **kwargs):
    from inspect import ismodule

    for arg in args:
        if ismodule(arg):
            ns_dict[arg.__name__] = arg
        elif isinstance(arg, str):
            exec(arg, ns_dict)
        else:
            ns_dict.update(arg)

    ns_dict.update(kwargs)




import parametrize_from_file
from vector import Vector, dot
from pytest import approx

@parametrize_from_file
def test_dot(a, b, expected):
    a, b = Vector(*a), Vector(*b)
    assert dot(a, b) == approx(expected)




#!/usr/bin/env python3

import json, toml, yaml, nestedtext as nt
import functools

@functools.wraps(json.load)
def _load_json(path):
    with open(path) as f:
        return json.load(f)

@functools.wraps(yaml.safe_load)
def _load_yml(path):
    with open(path) as f:
        return yaml.safe_load(f)

_LOADERS = {
        '.json': _load_json,
        '.yaml': _load_yml,
        '.yml': _load_yml,
        '.toml': toml.load,
        '.nt': nt.load,
}

def add_loader(suffix, loader):
    """
    Read test parameters from a custom file type.

    Arguments:
        suffix (str):
            The file suffix to associate with the given loader.  This should 
            include a leading dot, e.g. ``'.yml'``.  It is possible to 
            overwrite the default suffixes with this function.

        loader (collections.abc.Callable):
            A function that will be used to load parameters from a file.  It 
            should have the following signature::

                def loader(path: pathlib.Path) -> Dict[str, List[Dict[str, Any]]]

            In other words, it should accept a path and return the top-level 
            data structure expected by :deco:`parametrize_from_file`.

    Note:
        Each file is only loaded once per pytest session.
    """
    _LOADERS[suffix] = loader

def drop_loader(suffix):
    """
    Prevent parameters from being read from files with the given suffix.
    """
    del _LOADERS[suffix]

def get_loaders():
    """
    Return the dictionary of known loaders.
    """
    return _LOADERS



import pytest, re
from .errors import ConfigError
from unittest.mock import MagicMock
from contextlib2 import nullcontext

def cast(**funcs):
    """
    Return a schema function that will apply functions to the given test 
    parameters.

    Arguments:
        funcs (dict):
            Keyword arguments where each key is the name of a parameter and 
            each value if a function to transform that parameter with.

    Any keys that are missing will be silently ignored; this makes it easier to 
    provide default values with `defaults`.  Note that if you use `defaults` 
    before `cast`, then the default values will be transformed as if they had 
    been read from the parameter file.  In contrast, if you use `cast` before 
    `defaults`, then the default values will be used directly.

    See the :doc:`/python_snippets` and :doc:`/optional_params` tutorials for 
    some good in-context examples.

    Example:

        >>> f = cast(a=int)
        >>> f({'a': '1'})
        {'a': 1}
        >>> f({})
        {}
    """

    def schema(params):
        for key, func in funcs.items():
            if key in params:
                params[key] = func(params[key])

        return params

    return schema

def defaults(**defaults):
    """
    Return a schema function that will add the given default values to test 
    cases.  In other words, this is an easy way to provide default values for 
    one or more test parameters.

    Arguments:
        defaults:
            Keyword arguments where each key is the name of a parameter and 
            each value is the corresponding default.

    See the :doc:`/optional_params` tutorial more much more info.

    Example:
        
        >>> f = defaults(a=0, b=0)
        >>> f({'b': 1})
        {'a': 0, 'b': 1}
    """
    return lambda params: {**defaults, **params}

def error(exc_spec, *, globals=None):
    """\
    Create a context manager that will check that the specified exception is
    raised.

    Arguments:
        exc_spec (str, list, dict):
            This argument specifies what kind of exception to expect (and 
            optionally how to check various aspects of it).

            If string: A string that evaluates to an exception type.  This can 
            also be ``"none"`` to specify that the returned context manager 
            should not expect any exception to be raised.

            If list: A list of the above strings that evaluate to exception 
            types.  The actual exception must be one of these types.

            If dict: The following keys are understood:

            - "type" (required): A string or list of strings that evaluate 
              to exception types.  The context manager will require that 
              an exception of one of the given types is raised.

            - "message" (optional): A string or list of strings that should 
              appear verbatim in the error message.

            - "pattern" (optional): A string or list of strings 
              representing patterns that should appear in the error 
              message.  Each string is interpreted as a regular expression, 
              in the same manner as the *match* argument to 
              `pytest.raises`.

            - "attrs" (optional): A dictionary of attributes that the 
              exception object should have.  The dictionary keys give the 
              attribute names, and should be strings.  The dictionary 
              values give the attribute values, and should also be strings.  
              Each value string will be evaluated in the context of the given  
              namespace to get the expected values.

            - "assertions" (optional): A string containing python code that 
              will be executed when the expected exception is detected.  Any 
              *globals* specified will be available to this code.  In addition, 
              the exception object itself will be available via the *exc* 
              variable.  This field is typically used to make free-form 
              assertions about the exception object.

            - "cause" (optional): An integer (or string that can be converted 
              to an integer) 'n'.  All other checks will apply to the n-th 
              direct cause of the caught exception, instead of the exception 
              itself.  This is useful in cases where the exception you want to 
              test is wrapped in some other exception.

            Note that string values are accepted for every one of these keys, 
            because this method is meant to help with parsing exception 
            information from a text file, e.g. in the NestedText_ format.  All 
            evaluations and executions are deferred for as long as possible.

        globals (dict):
            The global variables to use when evaluating/executing code related 
            to the exception.  This can be any kind of mapping, including 
            `Namespace`.  By default, only the built-in variables are 
            available.

    Returns:
        A context manager that can be used to check if the kind of 
        exception specified by *exc_spec* was raised.

    Examples:
        Using a built-in exception (so no need to specify a namespace) and not 
        checking the error message::

            >>> p = {'type': 'ZeroDivisionError'}
            >>> with error(p):
            ...    1/0

        Using a custom exception::

            >>> class MyError(Exception):
            ...     pass
            ...
            >>> with_err = Namespace(MyError=MyError)
            >>> p = {'type': 'MyError', 'pattern': r'\\d+'}
            >>> with with_err.error(p):
            ...    raise MyError('404')

    Details:
        The returned context manager is re-entrant, which makes it possible 
        to stack :deco:`parametrize_from_file` invocations that make use of 
        method (e.g. via the *schema* argument).
    """

    if exc_spec == 'none':
        return ExpectSuccess()

    err = ExpectError(globals=globals)

    if isinstance(exc_spec, str):
        err.type_str = exc_spec
    else:
        def require_list(x):
            return x if isinstance(x, list) else [x]

        err.type_str = exc_spec['type']
        err.messages = require_list(exc_spec.get('message', []))
        err.patterns = require_list(exc_spec.get('pattern', []))
        err.attr_strs = exc_spec.get('attrs', {})
        err.assertions_str = exc_spec.get('assertions', '')
        err.cause_str = exc_spec.get('cause', '')

    return err

def error_or(*expected, globals=None, param='error', mock_factory=MagicMock):
    """
    Return a schema function that will expect to be given either an error or 
    the specified set of expected values.

    Arguments:
        expected (str):
            Any test parameters that must be specified when an exception is 
            *not* expected.  These parameters will be replaced with mocks when 
            an exception is expected.

        globals (dict):
            The global variables to use when building the `error` context 
            manager, e.g. for evaluating the exception type.  This can be any 
            kind of mapping, including `Namespace`.

        param (str):
            The name of the test parameter specifying the expected exception.

        mock_factory (Callable):
            A no-argument callable that can be used to create mock expected 
            values when an exception is expected.

    The purpose of this schema is to make it easier to test functions that can 
    raise exceptions.  Towards this end, this schema understands two sets of 
    parameters: one that defines an expected outcome when no exceptions are 
    expected, and another that defines an expected exception.  These parameter 
    sets are mutually exclusive: you cannot specify both at once.

    By default, exceptions are specified using the *error* parameter.  (The 
    name of this parameter can be changed via the *param* argument.)  If this 
    parameter is found, it will be processed by |NS_error|.  That means: (i) it 
    should match the format expected by that method and (ii) it will be 
    converted into a context manager that can be used to detect whether the 
    expected exception was in fact raised.  All of the *expected* keys will be 
    added to the output dictionary, and will be set to |MagicMock| objects.  
    These mocks are pretty good at standing in for other objects, so if you 
    need to do some mild processing on the expected values before performing 
    the actual test, it will often just work.  When it doesn't, the *error* 
    parameter will be truthy if an exception is expected and falsey otherwise, 
    so you can use that information to skip this kind of processing as 
    necessary.

    If the *error* parameter is not given, the resulting output will contain an 
    *error* key set to a no-op context manager.  All of the *expected* keys 
    will be passed through unchanged.

    For more information, see the :doc:`/exceptions` tutorial.

    Example:
        This example is somewhat contrived, but it shows how to easily test a 
        function that raises a custom exception for certain inputs.
        
           >>> from parametrize_from_file import Namespace, error_or

           >>> class DontGreet(Exception):
           ...     pass
           ...
           >>> def greet(name):
           ...     if name == 'Bob':
           ...         raise DontGreet(name)
           ...     return f"Hello, {name}!"
           ...
           >>> with_greet = Namespace(globals())
           >>> schema = with_greet.error_or('expected')
           >>> p1 = schema({
           ...         'given': 'Alice',
           ...         'expected': 'Hello, Alice!',
           ... })
           >>> with p1['error']:
           ...    assert greet(p1['given']) == p1['expected']
           ...
           >>> p2 = schema({
           ...         'given': 'Bob',
           ...         'error': {
           ...             'type': 'DontGreet',
           ...             'message': 'Bob',
           ...         },
           ... })
           >>> with p2['error']:
           ...    assert greet(p2['given']) == p2['expected']
           ...
        
        Note that the with block is the exact same for both assertions, even 
        though one tests for an error and the other doesn't.  This is what 
        makes `error_or()` useful for writing parametrized tests.
    """

    error_key = param

    def schema(params):
        if error_key not in params:
            params[error_key] = ExpectSuccess()

        else:
            bad_keys = set(params) & set(expected)
            if bad_keys:
                err = ConfigError(
                        bad_keys=bad_keys,
                        error_key=error_key,
                )
                err.brief = "must specify either an expected value or an error, not both"
                err.info += lambda e: f"expected value parameter(s): {','.join(e.bad_keys)}"
                err.info += "error parameter: {error_key}"
                raise err

            params[error_key] = error(
                    params[error_key],
                    globals=globals,
            )
            for key in expected:
                params[key] = mock_factory()

        return params

    return schema

class ExpectSuccess(nullcontext):

    def __repr__(self):
        return f'<{self.__class__.__name__}>'

    def __bool__(self):
        return False

class ExpectError:

    # Normally I'd use `@contextmanager` to make a context manager like this, 
    # but generator-based context managers cannot be reused.  This is a problem 
    # for tests, because if a test using this context manager is parametrized, 
    # the same context manager instance will need to be reused multiple times.  
    # The only way to support this is to implement the context manager from 
    # scratch.

    def __init__(self, *, globals=None, type_str=Exception, messages=[], patterns=[], attr_strs={}, assertions_str='', cause_str=''):
        if globals is None:
            globals = {}

        self.globals = globals
        self.type_str = type_str
        self.messages = messages
        self.patterns = patterns
        self.attr_strs = attr_strs
        self.assertions_str = assertions_str
        self.cause_str = cause_str

    def __repr__(self):
        attrs = {
                'type': self.type_str,
                'messages': self.messages,
                'patterns': self.patterns,
                'attrs': self.attr_strs,
                'assertions': self.assertions_str,
                'cause': self.cause_str,
        }
        attr_str = ' '.join(
                f'{k}={v!r}'
                for k, v in attrs.items() if v
        )
        return f'<{self.__class__.__name__} {attr_str}>'

    def __bool__(self):
        return True

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, exc_tb):
        from .namespace import Namespace

        # Don't include this method in stack traces generated by pytest.
        __tracebackhide__ = True

        ns = Namespace(self.globals)
        type = ns.eval(self.type_str)
        if isinstance(type, list):
            type = tuple(type)

        assert exc_type is not None, f"DID NOT RAISE {type}"

        del exc_type, exc_tb
        for i in range(int(self.cause_str or 0)):
            assert exc_value.__cause__ is not None, f"{exc_value.__class__.__name__} has no direct cause"
            exc_value = exc_value.__cause__

        if not isinstance(exc_value, type):
            return False

        exc_str = str(exc_value)

        for msg in self.messages:
            assert msg in exc_str, f'{msg!r} not in {exc_str!r}'

        for pat in self.patterns:
            assert re.search(pat, exc_str), "regex pattern {pat!r} does not match {exc_str!r}"

        for attr, value_str in self.attr_strs.items():
            assert hasattr(exc_value, attr)
            assert getattr(exc_value, attr) == ns.eval(value_str)

        if self.assertions_str:
            ns.fork(exc=exc_value).exec(self.assertions_str)

        return True




import math

class Vector:

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def normalize(self):
        m = math.sqrt(self.x**2 + self.y**2)

        if m == 0:
            raise NullVectorError
        else:
            self.x /= m
            self.y /= m

class NullVectorError(Exception):
    pass



import parametrize_from_file
from pytest import approx
from parametrize_from_file import Namespace, cast

with_vec = Namespace('from vector import *')

@parametrize_from_file(
        schema=[
            cast(
                given=with_vec.eval,
                expected=with_vec.eval,
            ),
            with_vec.error_or('expected'),
        ],
)
def test_normalize(given, expected, error):
    with error:
        given.normalize()

        assert given.x == approx(expected.x)
        assert given.y == approx(expected.y)



#!/usr/bin/env python3

from collections.abc import Iterable

def is_iterable(obj, not_iterable=(str, bytes)):
    if not isinstance(obj, Iterable):
        return False
    if not_iterable and isinstance(obj, not_iterable):
        return False
    return True





import vector
import parametrize_from_file

from pytest import approx
from parametrize_from_file import Namespace

# Define this object globally, because it will be useful for many tests.
with_vec = Namespace('from vector import *')

@parametrize_from_file
def test_dot(a, b, expected):
    a, b, expected = with_vec.eval(a, b, expected)
    assert vector.dot(a, b) == approx(expected)



import vector
import parametrize_from_file
from parametrize_from_file import Namespace

with_vec = Namespace("from vector import *")

@parametrize_from_file
def test_to_vector(given, expected):
    given = with_vec.exec(given)['obj']
    converted = vector.to_vector(given)
    expected = with_vec.eval(expected)

    assert converted.x == expected.x
    assert converted.y == expected.y



#!/usr/bin/env python3

"""
Parametrize test functions with values read from config files.

Note that this module itself can be used as an alias for `parametrize`, e.g.:

.. code-block::

    import parametrize_from_file

    @parametrize_from_file
    def test_my_func(given, expected):
        assert my_func(given) == expected

This is the recommended usage if you do not need to import any other names from 
the module.  If you do need to import other names, the recommendation is 
instead to import the module as ``pff``, e.g.:

.. code-block::

    import parametrize_from_file as pff

    @pff.fixture
    def env(request):
        return request.param

    @pff.parametrize
    def test_my_func(given, expected, env):
        assert my_func(given) == expected
"""

from .parameters import parametrize, fixture, load_parameters
from .namespace import Namespace, star
from .schema import defaults, cast, error, error_or
from .loaders import add_loader, drop_loader
from .errors import ConfigError

__version__ = '0.15.0'

for obj in [parametrize, fixture, Namespace, star, defaults, cast, error, error_or, add_loader, drop_loader, load_parameters, ConfigError]:
    obj.__module__ = 'parametrize_from_file'

del obj

# Hack to make the module directly usable as a decorator.  Only works for 
# python 3.5 or higher.  See this Stack Overflow post:
# https://stackoverflow.com/questions/1060796/callable-modules

import sys
import functools

class CallableModule(sys.modules[__name__].__class__):

    def __call__(self, *args, **kwargs):
        return parametrize(*args, **kwargs)

sys.modules[__name__].__class__ = CallableModule
del sys, functools, CallableModule



class Vector:

    def __init__(self, x, y):
        self.x = x
        self.y = y

# start here
def to_vector(obj):
    if isinstance(obj, Vector):
        return obj

    # If the input object is a container with two elements, use those elements 
    # to construct a vector.

    try: return Vector(*obj)
    except: pass

    # If the input object has x and y attributes, use those attributes to 
    # construct a vector.
    
    try: return Vector(obj.x, obj.y)
    except: pass



import vector
import parametrize_from_file
from parametrize_from_file import Namespace, cast, error_or

with_py = Namespace()
with_vec = Namespace('from vector import *')

@parametrize_from_file(
        schema=[
            cast(expected=with_vec.eval),
            error_or('expected'),
        ],
        indirect=['tmp_files'],
)
def test_load(tmp_files, expected, error):
    with error:
        assert vector.load(tmp_files / 'vectors.txt') == expected



import math
import vector
import parametrize_from_file
from pytest import approx
from parametrize_from_file import Namespace, cast, defaults

with_math = Namespace('from math import *')
with_vec = with_math.fork('from vector import *')

@parametrize_from_file(
        schema=[
            cast(
                angle=with_math.eval,
                kwargs=with_math.eval,
                expected=with_vec.eval,
            ),
            defaults(
                kwargs={},
            ),
        ],
)
def test_from_angle(angle, kwargs, expected):
    actual = vector.from_angle(angle, **kwargs)

    assert actual.x == approx(expected.x)
    assert actual.y == approx(expected.y)



#!/usr/bin/env python3

import pytest, sys, os
from unittest.mock import Mock, MagicMock
from parametrize_from_file import Namespace, star
from operator import itemgetter

class IgnoreMissing:

    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return f'IgnoreMissing({self.items!r})'

    def __eq__(self, other):
        subset = {k: other[k] for k in self.items if k in other}
        return subset == self.items

parametrize_init_fork = pytest.mark.parametrize(
        'args, kwargs, expected', [

            # One name:
            ([{'a': 1}], {}, {'a': 1}),
            (['a = 1'], {}, {'a': 1}),
            ([sys], {}, {'sys': sys}),
            ([], {'a': 1}, {'a': 1}),

            # Two names:
            ([{'a': 1, 'b': 2}], {}, {'a': 1, 'b': 2}),
            ([{'a': 1}, {'b': 2}], {}, {'a': 1, 'b': 2}),
            ([{'a': 1}, 'b = 2'], {}, {'a': 1, 'b': 2}),
            ([{'a': 1}, sys], {}, {'a': 1, 'sys': sys}),
            ([{'a': 1}], {'b': 2}, {'a': 1, 'b': 2}),

            (['a, b = 1, 2'], {}, {'a': 1, 'b': 2}),
            (['a = 1', {'b': 2}], {}, {'a': 1, 'b': 2}),
            (['a = 1', 'b = 2'], {}, {'a': 1, 'b': 2}),
            (['a = 1', sys], {}, {'a': 1, 'sys': sys}),
            (['a = 1'], {'b': 2}, {'a': 1, 'b': 2}),

            ([sys, {'b': 2}], {}, {'sys': sys, 'b': 2}),
            ([sys, 'b = 2'], {}, {'sys': sys, 'b': 2}),
            ([sys, os], {}, {'sys': sys, 'os': os}),
            ([sys], {'b': 2}, {'sys': sys, 'b': 2}),

            ([{'b': 2}], {'a': 1}, {'a': 1, 'b': 2}),
            (['b = 2'], {'a': 1}, {'a': 1, 'b': 2}),
            ([sys], {'a': 1}, {'a': 1, 'sys': sys}),
            ([], {'a': 1, 'b': 2}, {'a': 1, 'b': 2}),

            # Order matters:
            ([{'a': 1}, {'a': 2}], {}, {'a': 2}),
            ([{'a': 1}, 'a = 2'], {}, {'a': 2}),
            ([{'a': 1}], {'a': 2}, {'a': 2}),

            (['a = 1', {'a': 2}], {}, {'a': 2}),
            (['a = 1', 'a = 2'], {}, {'a': 2}),
            (['a = 1'], {'a': 2}, {'a': 2}),
        ],
)

@parametrize_init_fork
def test_init(args, kwargs, expected):
    ns = Namespace(*args, **kwargs)
    for k, v in expected.items():
        assert ns[k] == v

def test_repr():
    ns = Namespace(a=1)
    assert repr(ns) == "Namespace({'a': 1})"

def test_operators():
    ns = Namespace(a=1)
    assert ns['a'] == 1
    assert len(ns) == 1
    assert list(ns.items()) == [('a', 1)]
    assert list(ns.keys()) == ['a']
    assert list(ns.values()) == [1]

@parametrize_init_fork
def test_copy(args, kwargs, expected):
    ns1 = Namespace(*args, **kwargs)
    ns2 = ns1.copy()
    for k, v in expected.items():
        assert ns2[k] == v

@parametrize_init_fork
def test_fork(args, kwargs, expected):
    ns = Namespace().fork(*args, **kwargs)
    for k, v in expected.items():
        assert ns[k] == v

def test_fork_overwrite():
    ns1 = Namespace(a=1, b=2)
    ns2 = ns1.fork(b=3, c=4)

    assert ns1 == {'a': 1, 'b': 2}
    assert ns2 == {'a': 1, 'b': 3, 'c': 4}

def test_fork_unpickleable():
    # Only globally-defined functions can be pickled.
    def unpickleable():
        pass

    ns1 = Namespace(a=unpickleable)
    ns2 = ns1.fork(b=2)

    assert ns1 == {'a': unpickleable}
    assert ns2 == {'a': unpickleable, 'b': 2}

def test_star_1():
    import mock_module
    assert star(mock_module) == {'a': 1}

def test_star_2():
    import mock_module_all
    assert star(mock_module_all) == {'b': 2}

@pytest.mark.parametrize(
        'globals, args, kwargs, invoke, expected', [
            # Basic data types:
            (
                {'a': 1},
                ['a + 1'],
                {},
                lambda f: f,
                2,
            ), (
                {'a': 1},
                ['a + 1', 'a + 2'],
                {},
                lambda f: f,
                [2, 3],
            ), (
                {'a': 1},
                [['a + 1']],
                {},
                lambda f: f,
                [2],
            ), (
                {'a': 1},
                [['a + 1', 'a + 2']],
                {},
                lambda f: f,
                [2, 3],
            ), (
                {'a': 1},
                [{'a': 'a + 1'}],
                {},
                lambda f: f,
                {'a': 2},
            ), (
                {'a': 1},
                [{'a': 'a + 1'}],
                {'keys': True},
                lambda f: f,
                {1: 2},
            ), (
                {'a': 1},
                [{'a': 'a + 1', 'b': 'a + 2'}],
                {},
                lambda f: f,
                {'a': 2, 'b': 3},
            ),

            # Recursion:
            (
                {'a': 1},
                [[['a + 1']]],
                {},
                lambda f: f,
                [[2]],
            ), (
                {'a': 1},
                [[{'a': 'a + 1'}]],
                {},
                lambda f: f,
                [{'a': 2}],
            ), (
                {'a': 1},
                [{'a': ['a + 1']}],
                {},
                lambda f: f,
                {'a': [2]},
            ), (
                {'a': 1},
                [{'a': ['a + 1']}],
                {'keys': True},
                lambda f: f,
                {1: [2]},
            ), (
                {'a': 1},
                [{'a': {'b': 'a + 1'}}],
                {},
                lambda f: f,
                {'a': {'b': 2}},
            ), (
                {'a': 1},
                [{'a': {'a+1': 'a+2'}}],
                {'keys': True},
                lambda f: f,
                {1: {2: 3}},
            ),

            # Deferral
            (
                {'a': 1},
                [],
                {},
                lambda f: f('a + 1'),
                2,
            ), (
                {'a': 1},
                ['a + 1'],
                {'defer': True},
                lambda f: f(),
                2,
            ), (
                {'a': 1},
                ['a + 1'],
                {'defer': True},
                lambda f: f.eval(),
                2,
            ), (
                {},
                ['1/0'],
                {'defer': True},
                lambda f: 'no error',
                'no error',
            ), (
                {'a': 1},
                [{'a': 'a + 1'}],
                {'defer': True, 'keys': True},
                lambda f: f(),
                {1: 2},
            ), (
                {'a': 1},
                [],
                {'defer': True},
                lambda f: f('a + 1')(),
                2,
            ), (
                {},
                [],
                {'defer': True},
                lambda f: f('1/0') and 'no error',
                'no error',
            ), (
                {'a': 1},
                [],
                {'defer': True, 'keys': True},
                lambda f: f({'a': 'a + 1'})(),
                {1: 2},
            ),
        ],
)
def test_eval(globals, args, kwargs, invoke, expected):
    ns = Namespace(globals)
    f = ns.eval(*args, **kwargs)
    assert invoke(f) == expected

def test_eval_immutable():
    ns = Namespace(a=1)
    assert ns == {'a': 1}

    ns.eval('a + 1')
    assert ns == {'a': 1}

@pytest.mark.parametrize(
        'mock', [Mock(), MagicMock()],
)
def test_eval_mock(mock):
    ns = Namespace()
    assert ns.eval(mock) is mock

@pytest.mark.parametrize(
        'globals, args, kwargs, error', [
            # Name errors:
            ({'a': 1}, ['b + 1'], {}, NameError),
            ({'a': 1}, ['a + 1', 'b + 2'], {}, NameError),
            ({'a': 1}, [['b + 1']], {}, NameError),
            ({'a': 1}, [{'a': 'b + 1'}], {}, NameError),
            ({'a': 1}, [{'b': '1'}], {'keys': True}, NameError),

            # Unhashable keys:
            ({'a': 1}, [{'[a]': '1'}], {'keys': True}, TypeError),
            ({'a': 1}, [{'{a: 1}': '1'}], {'keys': True}, TypeError),

            # Regular errors:
            ({}, ['1/0'], {}, ZeroDivisionError),
            ({}, ['[][1]'], {}, IndexError),
            ({}, ['{}["x"]'], {}, KeyError),

            # Not list/dict/str:
            ({'a': 1}, [1], {}, TypeError),
            ({'a': 1}, [('a + 1', 'a + 2')], {}, TypeError),
        ],
)
def test_eval_err(globals, args, kwargs, error):
    ns = Namespace(globals)
    with pytest.raises(error):
        ns.eval(*args, **kwargs)

@pytest.mark.parametrize(
        'globals, args, kwargs, invoke, expected', [
            (
                {'a': 1},
                ['b = a + 1'],
                {},
                lambda f: f,
                IgnoreMissing({'a': 1, 'b': 2}),
            ),

            # Deferral
            (
                {'a': 1},
                [],
                {},
                lambda f: f('b = a + 1'),
                IgnoreMissing({'a': 1, 'b': 2}),
            ), (
                {'a': 1},
                ['b = a + 1'],
                {'defer': True},
                lambda f: f(),
                IgnoreMissing({'a': 1, 'b': 2}),
            ), (
                {'a': 1},
                ['b = a + 1'],
                {'defer': True},
                lambda f: f.exec(),
                IgnoreMissing({'a': 1, 'b': 2}),
            ), (
                {},
                ['1/0'],
                {'defer': True},
                lambda f: 'no error',
                'no error',
            ), (
                {'a': 1},
                [],
                {'defer': True},
                lambda f: f('b = a + 1')(),
                IgnoreMissing({'a': 1, 'b': 2}),
            ), (
                {},
                [],
                {'defer': True},
                lambda f: f('1/0') and 'no error',
                'no error',
            ),

            # Name lookup
            (
                {'a': 1},
                ['b = a + 1'],
                {'get': 'b'},
                lambda f: f,
                2,
            ), (
                {'a': 1},
                [],
                {'get': 'b'},
                lambda f: f('b = a + 1'),
                2,
            ), (
                {'a': 1},
                [],
                {'get': ('a', 'b')},
                lambda f: f('b = a + 1'),
                (1, 2),
            ), (
                {'a': 1},
                [],
                {'get': ['a', 'b']},
                lambda f: f('b = a + 1'),
                (1, 2),
            ), (
                {'a': 1},
                [],
                {'get': itemgetter('b')},
                lambda f: f('b = a + 1'),
                2,
            ), (
                {'a': 1},
                [],
                {'get': 'b', 'defer': True},
                lambda f: f('b = a + 1')(),
                2,
            ), (
                {},
                [],
                {'get': 'b', 'defer': True},
                lambda f: f('1/0') and 'no error',
                'no error',
            ),
        ],
)
def test_exec(globals, args, kwargs, invoke, expected):
    ns1 = Namespace(globals)
    f = ns1.exec(*args, **kwargs)

    assert ns1 == globals
    assert invoke(f) == expected

def test_exec_immutable():
    ns1 = Namespace(a=1)
    ns2 = ns1.exec('b = a + 1')
    ns3 = ns2.exec('c = b + 1')

    assert ns1 == {'a': 1}

    # Can't simply compare to a dictionary anymore, because `exec()` adds a 
    # bunch of builtin names to the namespace.
    assert ns2['a'] == 1
    assert ns2['b'] == 2
    assert 'c' not in ns2

    assert ns3['a'] == 1
    assert ns3['b'] == 2
    assert ns3['c'] == 3

@pytest.mark.parametrize(
        'mock', [Mock(), MagicMock()],
)
def test_exec_mock(mock):
    ns = Namespace()
    assert ns.exec(mock) is mock

@pytest.mark.parametrize(
        'globals, src, kwargs, error', [
            ({'a': 1}, 'b', {}, NameError),
            ({'a': 1}, '1/0', {}, ZeroDivisionError),
            ({'a': 1}, 1, {}, TypeError),
            ({'a': 1}, 'b = a + 1', {'get': 'c'}, KeyError),
            ({'a': 1}, 'b = a + 1', {'get': itemgetter('c')}, KeyError),
        ],
)
def test_exec_err(globals, src, kwargs, error):
    ns = Namespace(globals)
    with pytest.raises(error):
        ns.exec(src, **kwargs)




class Vector:

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, v):
        return self.x == v.x and self.y == v.y

def load(path):
    vectors = []

    with open(path) as f:
        for i, line in enumerate(f):
            fields = line.split()
            if len(fields) != 2:
                raise IndexError(f"line {i+1}: expected 2 coordinates, got {len(fields)}")

            x, y = map(float, fields)
            v = Vector(x, y)
            vectors.append(v)

    return vectors



#!/usr/bin/env python3

import pytest
import parametrize_from_file as pff
from parametrize_from_file.schema import ExpectSuccess
from unittest.mock import MagicMock
from inspect import isclass

class MockError1(Exception):

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.__dict__.update(kwargs)

class MockError1a(MockError1):
    pass

class MockError2(Exception):
    pass

parametrize_get_cm = pytest.mark.parametrize(
        'get_cm', [
            lambda exc_spec, globals: pff.error(exc_spec, globals=globals),
            lambda exc_spec, globals: pff.Namespace(globals).error(exc_spec),
        ],
)
def wrap_error(exc_1, exc_2, cause=True):
    try:
        raise exc_1
    except Exception as err:
        if cause:
            raise exc_2 from err
        else:
            raise exc_2


def test_defaults():
    schema = pff.defaults(a=1, b=0)
    assert schema({'b': 2, 'c': 3}) == {'a': 1, 'b': 2, 'c': 3}

def test_cast():
    schema = pff.cast(a=int)
    assert schema({'a': '1', 'b': 2}) == {'a': 1, 'b': 2}

def test_cast_missing():
    schema = pff.cast(a=int)
    assert schema({}) == {}

@pytest.mark.parametrize(
        'globals, exc_spec, expected_errors, unexpected_errors', [
            # Type
            (
                {'E': MockError1},
                'none',
                [],
                [MockError2()],
            ), (
                {'E': MockError1},
                'E',
                [MockError1],
                [MockError2()],
            ), (
                {'E': MockError1},
                {'type': 'E'},
                [MockError1],
                [MockError2()],
            ), (
                {'E': MockError1},
                {'type': 'E'},
                [MockError1a],
                [MockError2()],
            ), (
                {'E': MockError1a},
                {'type': 'E'},
                [MockError1a],
                [MockError1()],
            ), (
                {'E1a': MockError1a, 'E2': MockError2},
                {'type': ['E1a', 'E2']},
                [MockError1a, MockError2],
                [MockError1()],
            ),

            # Message
            (
                {'E': MockError1},
                {'type': 'E', 'message': 'a'},
                [MockError1('a')],
                [MockError1('b')],
            ), (
                {'E': MockError1},
                {'type': 'E', 'message': ['a', 'b']},
                [MockError1('ab'), MockError1('ba')],
                [MockError1('a'), MockError1('b')],
            ),

            # Pattern
            (
                {'E': MockError1},
                {'type': 'E', 'pattern': r'\d{3}'},
                [MockError1('404')],
                [MockError1(r'\d{3}')],
            ), (
                {'E': MockError1},
                {'type': 'E', 'pattern': [r'\d{3}', r'\w{4}']},
                [MockError1('HTTP 404'), MockError1('404 HTTP')],
                [MockError1(r'\d{3}\w{4}'), MockError1('HTTP'), MockError1('404')],
            ),

            # Attrs
            (
                {'E': MockError1},
                {'type': 'E', 'attrs': {'a': '1'}},
                [MockError1(a=1)],
                [MockError1(), MockError1(a=2)],
            ), (
                {'E': MockError1},
                {'type': 'E', 'attrs': {'a': '1', 'b': '2'}},
                [MockError1(a=1, b=2)],
                [MockError1(a=1), MockError1(b=2), MockError1(a=1, b=1), MockError1(a=2, b=2)],
            ),

            # Assertions
            (
                {'E': MockError1},
                {'type': 'E', 'assertions': 'assert exc.f() == 1'},
                [MockError1(f=lambda: 1)],
                [MockError1(f=lambda: 2)],
            ),

            # Cause
            (
                {'E1': MockError1, 'E2': MockError2},
                {'type': 'E1', 'cause': 1},
                [lambda: wrap_error(MockError1, MockError2)],
                [
                    MockError1(), 
                    (MockError2, lambda: wrap_error(MockError1, MockError2, cause=False)),
                    (MockError1, lambda: wrap_error(MockError2, MockError1)),
                ],
            ), (
                {'E1': MockError1, 'E2': MockError2},
                {'type': 'E1', 'cause': '1', 'message': 'a'},
                [lambda: wrap_error(MockError1('a'), MockError1)],
                [
                    (MockError1, lambda: wrap_error(MockError1, MockError1('a'))),
                ],
            ), (
                {'E1': MockError1, 'E2': MockError2},
                {'type': 'E1', 'cause': '1', 'pattern': r'\d{3}'},
                [lambda: wrap_error(MockError1('404'), MockError1('HTTP'))],
                [
                    (MockError1, lambda: wrap_error(MockError1('HTTP'), MockError1('404'))),
                ],
            ), (
                {'E1': MockError1, 'E2': MockError2},
                {'type': 'E1', 'cause': '1', 'attrs': {'a': '1'}},
                [lambda: wrap_error(MockError1(a=1), MockError1(a=2))],
                [
                    (MockError1, lambda: wrap_error(MockError1(a=2), MockError1(a=1))),
                ],
            ),
        ],
)
@parametrize_get_cm
def test_error(globals, exc_spec, get_cm, expected_errors, unexpected_errors):
    cm = get_cm(exc_spec, globals)

    assert bool(cm) == bool(expected_errors)

    def unpack(err):
        if isinstance(err, tuple):
            return err
        else:
            return err.__class__, err

    def raise_or_call(err):
        if isinstance(err, Exception):
            raise err
        if isclass(err) and issubclass(err, Exception):
            raise err
        err()

    def noop():
        pass

    for err in expected_errors or [noop]:
        print("expected:", err)
        with cm:
            raise_or_call(err)

    for err_type, err in map(unpack, unexpected_errors):
        print("unexpected:", err_type, err)
        with pytest.raises((err_type, AssertionError)):
            with cm:
                raise_or_call(err)
    
@pytest.mark.parametrize(
        'globals, exc_spec, trigger_error, expected_error', [
            (
                {'E': MockError1},
                'F',
                MockError1,
                NameError,
            ), (
                {'E': MockError1},
                '1/0',
                MockError1,
                ZeroDivisionError,
            ), (
                {'E': MockError1}, 
                {'type': 'E', 'attrs': {'a': 'b'}},
                MockError1(a=1),
                NameError,
            ), (
                {'E': MockError1},
                {'type': 'E', 'assertions': 'a'},
                MockError1,
                NameError,
            ),
        ],
)
@parametrize_get_cm
def test_error_err(globals, exc_spec, get_cm, trigger_error, expected_error):
    with pytest.raises(expected_error):
        with get_cm(exc_spec, globals):
            raise trigger_error

@pytest.mark.parametrize(
        'globals, exc_spec, expected', [
            (
                {'E': MockError1},
                'none',
                "<ExpectSuccess>",
            ), (
                {'E': MockError1},
                'E',
                "<ExpectError type='E'>",
            ), (
                {'E': MockError1},
                {'type': 'E', 'message': 'a'},
                "<ExpectError type='E' messages=['a']>",
            ), (
                {'E': MockError1},
                {'type': 'E', 'pattern': 'a'},
                "<ExpectError type='E' patterns=['a']>",
            ), (
                {'E': MockError1},
                {'type': 'E', 'attrs': {'a': '1'}},
                "<ExpectError type='E' attrs={'a': '1'}>",
            ), (
                {'E': MockError1},
                {'type': 'E', 'assertions': 'assert True'},
                "<ExpectError type='E' assertions='assert True'>",
            ),
        ],
)
def test_error_repr(globals, exc_spec, expected):
    cm = pff.error(exc_spec, globals=globals)
    assert repr(cm) == expected

def test_error_or_expected():
    schema = pff.error_or('a')
    params = schema({'a': 1})

    assert set(params) == {'a', 'error'}
    assert params['a'] == 1
    assert isinstance(params['error'], ExpectSuccess)
    assert not params['error']

def test_error_or_error():
    schema = pff.error_or('a')
    params = schema({'error': 'ZeroDivisionError'})

    assert set(params) == {'a', 'error'}
    assert isinstance(params['a'], MagicMock)
    assert params['error']
    with params['error']:
        raise ZeroDivisionError

@pytest.mark.parametrize(
        'wrap_globals', [
            lambda x: x,
            lambda x: pff.Namespace(x),
        ],
)
def test_error_or_globals(wrap_globals):
    globals = wrap_globals({'E1': MockError1})
    schema = pff.error_or('a', globals=globals)
    params = schema({'error': 'E1'})

    with params['error']:
        raise MockError1

def test_error_or_param():
    schema = pff.error_or('a', param='err')
    params = schema({'err': 'ZeroDivisionError'})

    with params['err']:
        raise ZeroDivisionError

def test_error_or_mock_factory():
    sentinel = object()
    schema = pff.error_or('a', mock_factory=lambda: sentinel)
    params = schema({'error': 'ZeroDivisionError'})

    assert params['a'] is sentinel

def test_error_or_ambiguous():
    schema = pff.error_or('a')

    with pytest.raises(pff.ConfigError) as err:
        schema({'a': 1, 'error': 'ZeroDivisionError'})

    assert err.match(r"must specify either an expected value or an error, not both")
    assert err.match(r"expected value parameter\(s\): a")
    assert err.match(r"error parameter: error")

